"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var chai_1 = require("chai");
var Logger_1 = require("./Logger");
var chalk_1 = __importDefault(require("chalk"));
var sinon_1 = require("sinon");
var sinon = sinon_1.createSandbox();
describe('Logger', function () {
    var logger;
    var levels = [
        'error',
        'warn',
        'log',
        'info',
        'debug',
        'trace'
    ];
    beforeEach(function () {
        logger = new Logger_1.Logger(Logger_1.LogLevel.trace);
        sinon.restore();
        //disable chalk colors for testing
        sinon.stub(chalk_1.default, 'grey').callsFake(function (arg) { return arg; });
    });
    it('noop does nothing', function () {
        Logger_1.noop();
    });
    it('loglevel setter converts string to enum', function () {
        logger.logLevel = 'error';
        chai_1.expect(logger.logLevel).to.eql(Logger_1.LogLevel.error);
        logger.logLevel = 'info';
        chai_1.expect(logger.logLevel).to.eql(Logger_1.LogLevel.info);
    });
    it('uses LogLevel.log by default', function () {
        logger = new Logger_1.Logger();
        chai_1.expect(logger.logLevel).to.eql(Logger_1.LogLevel.log);
    });
    describe('log methods call correct error type', function () {
        it('error', function () {
            var stub = sinon.stub(logger, 'writeToLog').callsFake(function () { });
            logger.error();
            chai_1.expect(stub.getCalls()[0].args[0]).to.eql(console.error);
        });
        it('warn', function () {
            var stub = sinon.stub(logger, 'writeToLog').callsFake(function () { });
            logger.warn();
            chai_1.expect(stub.getCalls()[0].args[0]).to.eql(console.warn);
        });
        it('log', function () {
            var stub = sinon.stub(logger, 'writeToLog').callsFake(function () { });
            logger.log();
            chai_1.expect(stub.getCalls()[0].args[0]).to.eql(console.log);
        });
        it('info', function () {
            var stub = sinon.stub(logger, 'writeToLog').callsFake(function () { });
            logger.info();
            chai_1.expect(stub.getCalls()[0].args[0]).to.eql(console.info);
        });
        it('debug', function () {
            var stub = sinon.stub(logger, 'writeToLog').callsFake(function () { });
            logger.debug();
            chai_1.expect(stub.getCalls()[0].args[0]).to.eql(console.debug);
        });
        it('trace', function () {
            var stub = sinon.stub(logger, 'writeToLog').callsFake(function () { });
            logger.trace();
            chai_1.expect(stub.getCalls()[0].args[0]).to.eql(console.trace);
        });
    });
    it('skips all errors on error level', function () {
        logger.logLevel = Logger_1.LogLevel.off;
        var stub = sinon.stub(logger, 'writeToLog').callsFake(function () { });
        logger.trace();
        logger.debug();
        logger.info();
        logger.log();
        logger.warn();
        logger.error();
        chai_1.expect(stub.getCalls().map(function (x) { return x.args[0]; })).to.eql([]);
    });
    it('does not skip when log level is high enough', function () {
        logger.logLevel = Logger_1.LogLevel.trace;
        var stub = sinon.stub(logger, 'writeToLog').callsFake(function () { });
        logger.trace();
        logger.debug();
        logger.info();
        logger.log();
        logger.warn();
        logger.error();
        chai_1.expect(stub.getCalls().map(function (x) { return x.args[0]; })).to.eql([
            console.trace,
            console.debug,
            console.info,
            console.log,
            console.warn,
            console.error
        ]);
    });
    describe('time', function () {
        it('calls action even if logLevel is wrong', function () {
            logger.logLevel = Logger_1.LogLevel.error;
            var spy = sinon.spy();
            logger.time(Logger_1.LogLevel.info, null, spy);
            chai_1.expect(spy.called).to.be.true;
        });
        it('runs timer when loglevel is right', function () {
            logger.logLevel = Logger_1.LogLevel.log;
            var spy = sinon.spy();
            logger.time(Logger_1.LogLevel.log, null, spy);
            chai_1.expect(spy.called).to.be.true;
        });
        it('returns value', function () {
            logger.logLevel = Logger_1.LogLevel.log;
            var spy = sinon.spy(function () {
                return true;
            });
            chai_1.expect(logger.time(Logger_1.LogLevel.log, null, spy)).to.be.true;
            chai_1.expect(spy.called).to.be.true;
        });
        it('gives callable pause and resume functions even when not running timer', function () {
            logger.time(Logger_1.LogLevel.info, null, function (pause, resume) {
                pause();
                resume();
            });
        });
        it('waits for and returns a promise when a promise is returned from the action', function () {
            chai_1.expect(logger.time(Logger_1.LogLevel.info, ['message'], function () {
                return Promise.resolve();
            })).to.be.instanceof(Promise);
        });
    });
});
//# sourceMappingURL=Logger.spec.js.map